/*-*- mode:c;indent-tabs-mode:nil;c-basic-offset:2;tab-width:8;coding:utf-8 -*-│
│vi: set net ft=c ts=2 sts=2 sw=2 fenc=utf-8                                :vi│
╞══════════════════════════════════════════════════════════════════════════════╡
│ Copyright 2020 Justine Alexandra Roberts Tunney                              │
│                                                                              │
│ Permission to use, copy, modify, and/or distribute this software for         │
│ any purpose with or without fee is hereby granted, provided that the         │
│ above copyright notice and this permission notice appear in all copies.      │
│                                                                              │
│ THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL                │
│ WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED                │
│ WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE             │
│ AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL         │
│ DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR        │
│ PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER               │
│ TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR             │
│ PERFORMANCE OF THIS SOFTWARE.                                                │
╚─────────────────────────────────────────────────────────────────────────────*/
#include "libc/assert.h"
#include "libc/bits/weaken.h"
#include "libc/macros.internal.h"
#include "libc/mem/mem.h"
#include "libc/nexgen32e/bsf.h"
#include "libc/nexgen32e/bsr.h"
#include "libc/runtime/runtime.h"
#include "libc/sysv/errfuns.h"

static struct CxaAtexitBlocks {
  struct CxaAtexitBlock {
    unsigned mask;
    struct CxaAtexitBlock *next;
    struct CxaAtexit {
      void *fp;
      void *arg;
      void *pred;
    } p[ATEXIT_MAX];
  } * p, root;
} __cxa_blocks;

/**
 * Adds global destructor.
 *
 * Destructors are called in reverse order. They won't be called if the
 * program aborts or _exit() is called. Invocations of this function are
 * usually generated by the C++ compiler. Behavior is limitless if some
 * other module has linked calloc().
 *
 * @param fp is void(*)(T)
 * @param arg is passed to callback
 * @param pred can be non-null for things like dso modules
 * @return 0 on success or nonzero w/ errno
 * @note folks have forked libc in past just to unbloat atexit()
 */
noasan int __cxa_atexit(void *fp, void *arg, void *pred) {
  unsigned i;
  struct CxaAtexitBlock *b, *b2;
  b = __cxa_blocks.p;
  if (!b) b = __cxa_blocks.p = &__cxa_blocks.root;
  if (!~b->mask) {
    if (weaken(calloc) &&
        (b2 = weaken(calloc)(1, sizeof(struct CxaAtexitBlock)))) {
      b2->next = b;
      __cxa_blocks.p = b = b2;
    } else {
      return enomem();
    }
  }
  i = bsr(~b->mask);
  assert(i < ARRAYLEN(b->p));
  b->mask |= 1u << i;
  b->p[i].fp = fp;
  b->p[i].arg = arg;
  b->p[i].pred = pred;
  return 0;
}

/**
 * Triggers global destructors.
 *
 * They're called in LIFO order. If a destructor adds more destructors,
 * then those destructors will be called immediately following, before
 * iteration continues.
 *
 * @param pred can be null to match all
 */
noasan void __cxa_finalize(void *pred) {
  unsigned i;
  unsigned long mask;
  struct CxaAtexitBlock *b, *b2;
StartOver:
  if ((b = __cxa_blocks.p)) {
    for (;;) {
      mask = b->mask;
      while (mask) {
        i = bsf(mask);
        mask &= ~(1u << i);
        if (!pred || pred == b->p[i].pred) {
          b->mask &= ~(1u << i);
          if (b->p[i].fp) {
            ((void (*)(void *))b->p[i].fp)(b->p[i].arg);
            goto StartOver;
          }
        }
      }
      if (!pred) {
        b2 = b->next;
        if (b2) {
          assert(b != &__cxa_blocks.root);
          if (weaken(free)) {
            weaken(free)(b);
          }
        }
        __cxa_blocks.p = b2;
        goto StartOver;
      } else {
        if (b->next) {
          b = b->next;
        } else {
          break;
        }
      }
    }
  }
}
